@@ -161,7 +161,7 @@ class MySQLParser(SQLParser):
     @_('from_table AS identifier')
     def from_table(self, p):
         entity = p.from_table
-        entity.alias = str(p.identifier)
+        entity.alias = p.identifier
         return entity
 
     @_('LPAREN query RPAREN')
@@ -212,7 +212,7 @@ class MySQLParser(SQLParser):
         col = p.result_column
         if col.alias:
             raise ParsingException(f'Attempt to provide two aliases for {str(col)}')
-        col.alias = str(p.identifier)
+        col.alias = p.identifier
         return col
 
     @_('LPAREN select RPAREN')
@@ -267,9 +267,9 @@ class MySQLParser(SQLParser):
     def expr_list_or_nothing(self, p):
         pass
 
-    @_('CAST LPAREN expr AS identifier RPAREN')
+    @_('CAST LPAREN expr AS ID RPAREN')
     def expr(self, p):
-        return TypeCast(arg=p.expr, type_name=str(p.identifier))
+        return TypeCast(arg=p.expr, type_name=str(p.ID))
 
     @_('enumeration')
     def expr_list(self, p):
