@@ -47,10 +47,11 @@ class JoinStep(PlanStep):
 
 class UnionStep(PlanStep):
     """Union of two dataframes, producing a new dataframe"""
-    def __init__(self, left, right, *args, **kwargs):
+    def __init__(self, left, right, unique, *args, **kwargs):
         super().__init__(*args, **kwargs)
         self.left = left
         self.right = right
+        self.unique = unique
 
 
 class OrderByStep(PlanStep):
@@ -88,7 +89,6 @@ class ApplyPredictorStep(PlanStep):
         self.dataframe = dataframe
 
 
-
 class ApplyPredictorRowStep(PlanStep):
     """Applies a mindsdb predictor to one row of values and returns a dataframe of one row, the predictor."""
     def __init__(self, namespace, predictor, row_dict, *args, **kwargs):
@@ -98,4 +98,20 @@ class ApplyPredictorRowStep(PlanStep):
         self.row_dict = row_dict
 
 
+class MapReduceStep(PlanStep):
+    """Applies a step for each value in a list, and then reduces results to a single dataframe"""
+    def __init__(self, values, step, reduce, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self.values = values
+        self.step = step
+        self.reduce = reduce
+
+
+class MultipleSteps(PlanStep):
+    def __init__(self, steps, reduce, *args, **kwargs):
+        """Runs multiple steps and reduces results to a single dataframe"""
+        super().__init__(*args, **kwargs)
+        self.steps = steps
+        self.reduce = reduce
+
 
